import struct, sys, time, logging, traceback
import socket
import re

from .GandanMsg import *
from .MMAP  import *

class GandanSub:
	def __init__(self, ip, port, _cmd, _io, path=None): # _io is wating time 
		if path == None:
			self.type_    = 0
			self.ip_port_ = (ip, port)
			self.io_	  = _io
			self.cmd_	 = _cmd
			self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.s = GandanSub.connect(None, self.s, self.ip_port_, self.cmd_, self.io_)
		else:
			size = 5000*20000
			item_size = 20000
			self.type_    = 1
			self.path_    = path
			self.cmd_     = _cmd
			self.topic    = _cmd.split('_')[1]
			#open shm
			self.mmap     = MMAP(path+"/"+self.topic, size, item_size)

	def sub_shm(self, cb, obj=None, hb_cb=None):
		msg_list = self.mmap.readp()
		#print(len(msg_list))
		ret = ""
		for msg in msg_list:
			if obj == None:
				ret = cb(str(msg, 'utf-8'))
			else:
				ret = cb(str(msg,'utf-8'), obj)
		return ret

	def sub(self, cb, obj=None, hb_cb=None):
		_h = None
		if self.type_ == 0:
			try:
				_h = GandanMsg.recv(None, self.s)

				if _h.dat_.strip() == "HB":
					#logging.info("------------------HB----------------[%s]" % self.cmd_)
					if hb_cb == None:
						return 1
					if hb_cb(obj, _h) == -1:
						raise Exception("conn")
				else:
					if obj == None:
						return cb(_h)
					return cb(_h, obj)

				return 1

			except Exception as e:
				if not str(e) in ["timeout"]:
					_type, _value, _traceback = sys.exc_info()
					logging.info("#Error" + str(_type) + str(_value))
					for _err_str in traceback.format_tb(_traceback):
						logging.info(_err_str)
					try:
						logging.info(str(_h).strip())
					except Exception as er:
						logging.info(str(er))

				if str(e) in ["timeout"]:
					if hb_cb != None and hb_cb(obj, None) == -1:
						e = Exception("conn")
					return None
				elif str(e) in ["convert"]:
					return -2

				if str(e) in ["conn"]:
					logging.info(str(e)+": #Error Try to reconnect")
					display_cnt = 0
					while True:
						try:
							self.s.close()
							logging.info("#SUCC : previous socket close")
						except Exception as e:
							logging.info("#FAIL : previous socket close")
						finally:
							try:
								self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
								self.s = GandanSub.connect(None, self.s, self.ip_port_, self.cmd_, self.io_)
								logging.info("#SUCC : new socket connect")
								break
							except Exception as e:
								if display_cnt % 100 == 0:
									logging.info("#FAIL : new socket connect")
							finally:
								display_cnt = display_cnt + 1
								time.sleep(0.1)
		elif self.type_ == 1:
			return self.sub_shm(cb, obj, hb_cb)

	def close(self):
		if self.type_ == 0:
			self.s.close()
		elif self.type_ == 1:
			self.mmap.f.close()

	@staticmethod
	def connect(self, _sock, _ip_port, _cmd, _io):
		_sock.connect(_ip_port)
		_sock.settimeout(_io)
		_h = GandanMsg(_cmd, "INIT")
		_sock.send(bytes(_h))
		return _sock

def callback(_h):
	print(_h)

if __name__ == "__main__":
	_h = GandanSub("127.0.0.1", 8888, sys.argv[1])
	_h.sub(callback)



import os,struct,threading,time,gc,logging
from mmap import mmap
import fcntl
import datetime

class MMAP:
    (CNT_START, CNT_SIZE) = 3, 4
    (RSTAMP_START, RSTAMP_SIZE) = 7, 8
    (WSTAMP_START, WSTAMP_SIZE) = 15, 8
    (R_START, R_SIZE) = 23, 4
    (W_START, W_SIZE) = 27, 4
    OFFSET_START = 31

    def __init__(self, path, size, item_size, opt = []):
        (self.use_cnt, self.use_tstamp) = (0, 0)
        if "TSTAMP" in opt:
            self.use_tstamp = 1
        
        self.f = self.init_file(path, size)
        
        self.f.seek(0, os.SEEK_SET); self.m = mmap(self.f.fileno(), MMAP.OFFSET_START+size)
        self.countp()
        self.size      = size
        self.item_size = item_size
        self.item_num  = int(size/item_size)
        self.erase_buff_ = ''
        for i in range(0, item_size):
            self.erase_buff_ += ' '
        self.erase_buff_ = bytes(self.erase_buff_, "ascii")


    def init_file(self, path, size):
        if os.path.isfile(path):
            f = open(path, "rb+");f.seek(0, os.SEEK_END)
            filesize = f.tell() ;f.seek(0, os.SEEK_SET)

            if f.read(3).decode("ascii") == "QUE":
               if filesize < MMAP.OFFSET_START + size:
                   f.seek(filesize, os.SEEK_SET)
                   for i in range(filesize, MMAP.OFFSET_START+ size):
                       f.write(' '.encode("ascii"))
               return f
        f = open(path, "wb+");
        f.seek(0, os.SEEK_SET)
        f.write("QUE".encode("ascii"))

        f.seek(MMAP.CNT_START, os.SEEK_SET)
        f.write(struct.pack("i", 1))

        if self.use_tstamp:
            f.seek(MMAP.RSTAMP_START, os.SEEK_SET)
            f.write(struct.pack("d", 0.0))
            f.seek(MMAP.WSTAMP_START, os.SEEK_SET)
            f.write(struct.pack("d", 0.0))

        f.seek(MMAP.R_START, os.SEEK_SET)
        f.write(struct.pack("i", 0))
        f.seek(MMAP.W_START, os.SEEK_SET)
        f.write(struct.pack("i", 0))

        for i in range(0, MMAP.OFFSET_START + size):
            f.write(' '.encode("ascii"))
        return f

    def close(self):
        self.countm()
        self.m.close()
        self.f.close()

    def r (self): return struct.unpack("i", self.m[MMAP.R_START: MMAP.R_START + MMAP.R_SIZE])[0]
    def w (self): return struct.unpack("i", self.m[MMAP.W_START: MMAP.W_START + MMAP.W_SIZE])[0]
    def rp(self): self.m.seek(MMAP.R_START, os.SEEK_SET); self.m.write(struct.pack("i", self.r()+1))
    def wp(self): self.m.seek(MMAP.W_START, os.SEEK_SET); self.m.write(struct.pack("i", self.w()+1))
    def rs(self, r): self.m.seek(MMAP.R_START, os.SEEK_SET); self.m.write(struct.pack("i", r))
    def ws(self, w): self.m.seek(MMAP.W_START, os.SEEK_SET); self.m.write(struct.pack("i", w))
    def rstamp(self): return struct.unpack("d", self.m[MMAP.RSTAMP_START: MMAP.RSTAMP_START + MMAP.RSTAMP_SIZE])[0]
    def wstamp(self): return struct.unpack("d", self.m[MMAP.WSTAMP_START: MMAP.WSTAMP_START + MMAP.WSTAMP_SIZE])[0]
    def count(self):  return struct.unpack("i", self.m[MMAP.CNT_START: MMAP.CNT_START + MMAP.CNT_SIZE])[0]

    def countp(self): 
        while fcntl.flock(self.f, fcntl.LOCK_EX) == False:
            time.sleep(0.001)
            continue
        try:
            self.m.seek(MMAP.CNT_START, os.SEEK_SET); self.m.write(struct.pack("i", self.count() + 1))
        except Exception as e:
            print(str(e))
        finally:
            fcntl.flock(self.f, fcntl.LOCK_UN)

    def countm(self): 
        while fcntl.flock(self.f, fcntl.LOCK_EX) == False:
            time.sleep(0.001)
            continue
        try:
            self.m.seek(MMAP.CNT_START, os.SEEK_SET); self.m.write(struct.pack("i", self.count() - 1))
        except Exception as e:
            print(str(e))
        finally:
            fcntl.flock(self.f, fcntl.LOCK_UN)

    def writep(self, data): 
        while fcntl.flock(self.f, fcntl.LOCK_EX) == False:
            time.sleep(0.001)
            continue
        try:
            self.write(data); self.ws(self.w()+1); del(data); 
            if self.use_tstamp:
                self.m.seek(MMAP.WSTAMP_START, os.SEEK_SET)
                self.m.write(struct.pack('d', datetime.datetime.now().timestamp()))
        except Exception as e:
            print(str(e))
        finally:
            fcntl.flock(self.f, fcntl.LOCK_UN)

        return self.r();

    def readp(self)       : 
        while fcntl.flock(self.f, fcntl.LOCK_EX) == False:
            time.sleep(0.001)
            continue
        try:
            _l = self.read(); self.rs(self.r() + len(_l)); 
            if self.use_tstamp:
                self.m.seek(MMAP.RSTAMP_START, os.SEEK_SET)
                self.m.write(struct.pack('d', datetime.datetime.now().timestamp()))
        except Exception as e:
            print(str(e))
        finally:
            fcntl.flock(self.f, fcntl.LOCK_UN)
            return _l

    def write(self, data):
        write_pos = MMAP.OFFSET_START + (self.w()*self.item_size) % self.size
        write_res = (write_pos - MMAP.OFFSET_START) % self.item_size
        #clear que item before write
        try:
            if write_res == 0 : 
                self.m.seek(write_pos, os.SEEK_SET); self.m.write(self.erase_buff_)
                self.m.seek(write_pos, os.SEEK_SET); self.m.write(data)
            else              : 
                self.m.seek(write_pos, os.SEEK_SET); self.m.write(self.erase_buff_[0:write_res])
                self.m.seek(MMAP.OFFSET_START, os.SEEK_SET); self.m.write(self.erase_buff_[write_res: self.item_size])
                self.m.seek(write_pos, os.SEEK_SET); self.m.write(data[0: write_res])
                self.m.seek(MMAP.OFFSET_START, os.SEEK_SET); self.m.write(data[write_res: self.item_size ])
        except Exception as e:
            logging.info("try to write ", write_pos, "@", self.size + 11)

    def read(self):
        try:
            _l = []
            for i in range(self.r(), self.w()):
                read_pos = MMAP.OFFSET_START + (i*self.item_size) % self.size
                read_res = (read_pos - MMAP.OFFSET_START) % self.item_size
                try:
                    if read_pos == MMAP.OFFSET_START + self.size :
                        _l.append(self.m[MMAP.OFFSET_START:MMAP.OFFSET_START+self.item_size])
                    elif read_res == 0 : 
                        _l.append(self.m[read_pos:read_pos + self.item_size])
                    else             :
                        data = ""
                        self.m.seek(read_pos, os.SEEK_SET); data = self.m.read(self.item_size - read_res)
                        self.m.seek(MMAP.OFFSET_START, os.SEEK_SET); data = data + self.m.read(read_res)
                        _l.append(data)
                except Exception as e:
                    logging.info("try to read", read_pos, "@", self.size + MMAP.OFFSET_START)
            return _l
        except Exception as e:
            logging.info(str(e))

class RoboMMAP(threading.Thread, MMAP):
    def __init__(self, path, topic, callback, size, item_size):
        threading.Thread.__init__(self)
        MMAP.__init__(self, path, size, item_size)
        self.topic    = topic
        self.time    = 0.001
        self.callback = callback

    def run (self)   : self.handle()

    def handle(self, flag=True):
        #put handler in while loop
        while True:
            try:
                #time.sleep(self.time)
                if self.r() < self.w() and self.callback != None:
                    _list = self.readp()
                    self.callback(self.topic, _list, self.r(), self.w())
                else:
                    logging.info("nothing to read")

                if flag == False:
                    for i in range(0, len(_list)):
                        logging.info(_list[i].strip())
                    break
            except Exception as e:
                logging.info("#Error:"+str(e))
                break



#--*--coding=utf-8--*--
import struct, sys, time, re
from ast import literal_eval
from datetime import datetime
import socket
import threading, logging
import traceback

from .GandanMsg import *
from .MMAP  import *

class Gandan:
	def __init__(self, ip_port, path, size, item_size, debug=False):
		self.ip_port = ip_port
		self.path    = path
		self.sz      = size
		self.isz     = item_size

		self.pub_topic = {}
		self.sub_topic = {}
		self.sub_topic_websocket = {}
		self.rlist     = []

		self.debug = debug

		self.pub_lock_dict = {}

	@staticmethod
	def setup_log(path, level = logging.DEBUG):
		l_format = '%(asctime)s:%(msecs)03d^%(levelname)s^%(funcName)20s^%(lineno)04d^%(message)s'
		d_format = '%Y-%m-%d^%H:%M:%S'
		logging.basicConfig(filename=path, format=l_format, datefmt=d_format,level=level)

	def handler(self, _req):
		p = True; h = None
		_pubsub = ""
		while(p):
			try:
				if self.debug:
					_st = datetime.now()
					h = GandanMsg.recv(None, _req)
					_et = datetime.now()
					logging.info("RECV_DATA : %s" % str(h).strip())
					logging.info("RECV_TIME : %d" % ((_et-_st).seconds*1000000 + (_et-_st).microseconds))
				else:
					h = GandanMsg.recv(None, _req)

				(_cmd, _msg)	  = (h.cmd_, h.dat_)
				[_pubsub, _topic] = _cmd.split("_")
			except Exception as e:
				Gandan.error_stack()
				if str(e) in ['convert']:
					continue
				p = False
				continue

			#Note : PUB/SUB가 등록 시 경쟁하는 상황이 발생가능
			if _pubsub == "PUB": 
				try:
					if not _topic in self.pub_lock_dict.keys():
						self.pub_lock_dict[_topic] = threading.Lock()
					self.pub_lock_dict[_topic].acquire()
					if self.debug:
						_st = datetime.now()
						p = self.pub(_req, _topic, _msg)
						_et = datetime.now()
						logging.info("PUB_TIME : %d" % ((_et-_st).seconds*1000000 + (_et-_st).microseconds))
					else:
						p = self.pub(_req, _topic, _msg)
				except Exception as e:
					Gandan.error_stack()
				finally:
					self.pub_lock_dict[_topic].release()

			if _pubsub == "SUB":
				try: 
					if self.debug:
						_st = datetime.now()
						p = self.sub(_req, _topic, _msg)
						_et = datetime.now()
						logging.info("PUB_TIME : %d" % ((_et-_st).seconds*1000000 + (_et-_st).microseconds))
					else:
						p = self.sub(_req, _topic, _msg)
				except Exception as e:
					Gandan.error_stack()


		try:
			if _pubsub == "PUB":
				_req.close()
		except Exception as e:
			logging.info("#Error socket close for %s" % str(_req))

	def pub(self, _req, _topic, _msg):
		try:
			_path = self.path +_topic
			_pub = None

			#if (_topic == "ORDER" or _topic == "OM") and _req.getsockname()[0] != '127.0.0.1':
			#	_req.close()
			#	return False

			if not _topic in self.pub_topic.keys(): 
				logging.info("ADD new PUB TOPIC[%s]" % _topic)
				self.pub_topic[_topic] = []
				self.pub_topic[_topic].append(RoboMMAP(_path, _topic, self.mon, self.sz, self.isz))
				_pub = self.pub_topic[_topic][-1]
				logging.info("PUB que path : %s monitor start,r[%d]w[%d]" % (_path, _pub.r(), _pub.w()) + ", TOPIC : %s" % _topic)
			else:
				_pub = self.pub_topic[_topic][-1]

			if Gandan.version() < 3:
				_pub.writep(bytes(_msg))
			else:
				_pub.writep(bytes(_msg,'utf-8'))

			_pub.handle(False)
		except Exception as e:
			if _pub == None:
				logging.info(str(e))
				logging.info(self.pub_topic)
				return False
			else:
				Gandan.error_stack()
				logging.info("#Error During PUB[%s], " % _topic + "[%s..]" % _msg.strip()[0:50] + " r[%d]w[%d]" % (_pub.r(), _pub.w()))
				return False

		logging.info("PUB[%s], " % _topic + "^%s^%d" % (_msg.strip(), len(_msg.strip())) + " r[%d]w[%d]" % (_pub.r(), _pub.w()))
		return True

	def sub(self, _req, _topic, _msg, _type = 0):
		try:
			if _type == 0:
				if not _topic in self.sub_topic.keys():
					self.sub_topic[_topic] = []
					self.sub_topic[_topic].append(_req)
					logging.info("_topic : %s is created for SUB" % _topic)
				else:
					self.sub_topic[_topic].append(_req)
					logging.info("_topic : %s is appended for SUB" % _topic)
			else:
				if not _topic in self.sub_topic_websocket.keys():
					self.sub_topic_websocket[_topic] = []
					self.sub_topic_websocket[_topic].append(_req)
					logging.info("_topic : %s is created for SUB, websocket" % _topic)
				else:
					self.sub_topic_websocket[_topic].append(_req)
					logging.info("_topic : %s is appended for SUB, websocket" % _topic)
		except Exception as e:
			return False
		return True

	def start(self):
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
			logging.info("------------ MW Gandan Version[%d] Start --------------" % Gandan.version())
			s.bind(self.ip_port)
			s.listen(30)
			#Todo : 소켓 당 쓰레드를 띄우는 대신 다중 IO로 변경
			while True:
				try:
					_r, _a = s.accept()
					self.rlist.append(_r)
					t = threading.Thread(target = self.handler, args = (_r, ))
					t.start()
				except Exception as e:
					logging.info("%s"%str(e)+":%s"%str(e))
					break

	def mon(self, _topic, _data, r, w):
		logging.info("[%s]" % str(type(_data)))
		logging.info("mon data for topic[%s]" % _topic)
		if not _topic in self.sub_topic.keys() and not _topic in self.sub_topic_websocket.keys():
			logging.info("no topic for data")
			return

		error_req_list  = []
		error_req_listw = []

		#http://effbot.org/pyfaq/what-kinds-of-global-value-mutation-are-thread-safe.htm
		if _topic in self.sub_topic.keys():
			for i, _req in enumerate(self.sub_topic[_topic]):
				for d in _data:
					try:
						if Gandan.version() < 3:
							GandanMsg.send(None, _req, "[%d][%d]DATA_" % (r, w) +_topic, str(d))
						else:
							GandanMsg.send(None, _req, "[%d][%d]DATA_" % (r, w) +_topic, str(d,'utf-8'))
					except Exception as e:
						error_req_list.append(_req)

			for _req in error_req_list:
				self.sub_topic[_topic].remove(_req)

		if _topic in self.sub_topic_websocket.keys():
			for i, _req in enumerate(self.sub_topic_websocket[_topic]):
				for d in _data:
					try:
						if Gandan.version() < 3:
							GandanMsg.send_websocket(None, _req, "[%d][%d]DATA_" % (r, w) +_topic, str(d))
						else:
							GandanMsg.send_websocket(None, _req, "[%d][%d]DATA_" % (r, w) +_topic, str(d,'utf-8'))
					except Exception as e:
						Gandan.error_stack()
						error_req_listw.append(_req)

		for _req in error_req_listw:
			self.sub_topic_websocket[_topic].remove(_req)

		if len(error_req_list) > 0:
			logging.info("#Error in SUB [%s] : %d request is removed" % (_topic, len(error_req_list)))

		if len(error_req_listw) > 0:
			logging.info("#Error in SUB [%s] : %d requestw is removed" % (_topic, len(error_req_listw)))

	@staticmethod
	def error_stack(stdout = False):
		_type, _value, _traceback = sys.exc_info()
		logging.info("#Error" + str(_type) + str(_value))
		for _err_str in traceback.format_tb(_traceback):
			if stdout == False:
				logging.info(_err_str)
			else:
				print(_err_str)

	@staticmethod
	def version():
		return int(re.sub('\.','',sys.version.split(' ')[0][0]))



#--*--coding:utf-8--*--
import struct, sys, time, re
from ast import literal_eval
import threading, logging, socket

class GandanMsg:
	def __init__(self, _cmd, _dat): 
		if GandanMsg.version() < 3:
			self.total_size = 12 + len(bytes(_cmd)) + len(bytes(_dat))
		else:
			self.total_size = 12 + len(bytes(_cmd, "utf-8")) + len(bytes(_dat,"utf-8"))
		(self.sep_, self.cmd_, self.dat_) = (b'#', _cmd, _dat)

	def __bytes__(self):
		_b =  self.sep_
		_b += struct.pack("!i", self.total_size-4)
		if GandanMsg.version() < 3:
			_b += struct.pack("!i", len(bytes(self.cmd_))) + bytes(self.cmd_)
			_b += struct.pack("!i", len(bytes(self.dat_))) + bytes(self.dat_)
		else:
			_b += struct.pack("!i", len(bytes(self.cmd_, "utf-8"))) + bytes(self.cmd_, "utf-8")
			_b += struct.pack("!i", len(bytes(self.dat_, "utf-8"))) + bytes(self.dat_, "utf-8")
		return _b
	
	def __str__(self):
		return str(self.total_size)+":"+self.cmd_+":"+self.dat_

	@staticmethod
	def version():
		return int(re.sub('\.','',sys.version.split(' ')[0][0]))

	@staticmethod
	def conv2(self, _b):
		_c_sz = struct.unpack("!i", _b[:4])[0]; _b = _b[4:]
		_c	= str(_b[:_c_sz]); _b = _b[_c_sz:]
		_d_sz = struct.unpack("!i", _b[:4])[0]; _b = _b[4:]
		_d	  = str(_b[:_d_sz])	 
		return GandanMsg(_c, _d)

	@staticmethod
	def conv3(self, _b):
		_c_sz = struct.unpack("!i", _b[:4])[0]; _b = _b[4:]
		_c	= str(_b[:_c_sz], 'utf-8'); _b = _b[_c_sz:]
		_d_sz = struct.unpack("!i", _b[:4])[0]; _b = _b[4:]
		_d	  = str(_b[:_d_sz], 'utf-8')	 
		return GandanMsg(_c, _d)

	@staticmethod
	def send(self, _sock, _cmd, _msg):
		_h = GandanMsg(_cmd, _msg)
		if GandanMsg.version() < 3:
			_sock.send(_h.__bytes__())
		else:
			_sock.send(bytes(_h))

	@staticmethod
	def send_websocket(self, _sock, _cmd, _msg):
		#안그래도 느린넘한테 보낼 때는 strip을 해서보냄
		#data = bytearray(_msg.strip().encode('utf-8'))
		data = bytearray(_msg.encode('utf-8'))
		if len(data) > 126:
			data = bytearray([ord(b'\x81'), 126]) + bytearray(struct.pack('>H', len(data))) + data
			#logging.info("> %s" % data)
		else:
			data = bytearray([ord(b'\x81'), len(data)]) + data
			#logging.info("%s" % data)
		_sock.send(data)

	@staticmethod
	def recv(self, _sock):
		_b = ''; _sz = 0;
		while True:
			try:
				_b = _sock.recv(1)
			except socket.timeout as e:
				raise Exception('timeout')

			if len(_b) == 0:
				_sock.close()
				raise Exception('conn')
			
			if _b == b'#': 
				break
		try:
			_b   = _sock.recv(4)  
		except socket.timeout as e:
			raise Exception('timeout')

		try:
			_sz  = struct.unpack("!i", _b)[0]
			#logging.info("size of recv packet from socket => [%d]" % _sz)
		except Exception as e:
			raise Exception('convert')

		try:
			_b   = _sock.recv(_sz)
		except Exception as e:
			raise Exception('timeout')

		if len(_b) == 0:
			raise Exception('conn')
		try:
			if GandanMsg.version() < 3:
				return GandanMsg.conv2(None, _b)
			else:
				return GandanMsg.conv3(None, _b)
		except Exception as e:
			logging.info(str(e))
			raise Exception('convert')



#--*--encoding:utf-8--*--
import struct, sys
import socket
import re
import threading, time
import logging

from .GandanMsg import *
from .MMAP  import *

class GandanPub:
	def __init__(self, ip, port, path=None, hb_time = 30):
		if path == None:
			self.type_ = 0
			self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.ip_port_ = (ip, port)
			self.s = GandanPub.connect(None, self.s, self.ip_port_)
			self.prev_cmd_ = None
			self.prev_data_ = None
			self.cmd_ = None
			#for multiple use
			self.publock = threading.Lock()
			self.hb_flag = True
			self.hb_time = hb_time
		else:
			self.type_ = 1
			self.mmap = None
			self.path = path
		
	def hb(self, obj):
		while obj.hb_flag:
			try:
				self.pub(self.cmd_, "HB")
				#logging.info("-------------------HB-----------------[%s, %s]...done" % (self.cmd_, self.hb_flag))
				time.sleep(self.hb_time)
			except Exception as e:
				time.sleep(1)
				continue

	def pub_shm(self, _cmd, _data):
		if self.mmap == None:
			size = 5000*20000
			item_size = 20000
			self.topic = _cmd.split('_')[1]
			self.mmap = MMAP(self.path+"/"+self.topic, size, item_size)
		else:
			self.mmap.writep(bytes(_data, 'utf-8'))
		
	def pub(self, _cmd, _data):
		if self.type_ == 0:
			try:
				if self.cmd_ == None:
					self.cmd_ = _cmd
					self.hb_thr = threading.Thread(target=self.hb, args=(self,))
					self.hb_thr.start()
				try:
					self.publock.acquire()
					GandanMsg.send(None, self.s, _cmd, _data)
					#socket closing 없이 죽은 경우, 첫번째 전송까지는 Exception 발생하지 않음
					#적절한 수준의 전송보장은 아닌 것 같음
					self.prev_cmd_ = _cmd
					self.prev_data_ = _data
				except Exception as e:
					logging.info(str(e))
					raise Exception("connection lost")
				finally:
					self.publock.release()

			except Exception as e:
				#while True:
				#	try:
				#		self.publock.acquire()
				#		logging.info("connection lost retry ... ")
				#		self.s.close()
				#		self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
				#		self.s = GandanPub.connect(None, self.s, self.ip_port_)
				#		logging.info("connection lost retry ... SUCC ")

				#		if False:#self.prev_cmd_ != None and self.prev_data_ != None:
				#			self.pub(self.prev_cmd_, self.prev_data_)
				#			logging.info("send[%s][%s] .. Done" % (self.prev_cmd_, self.prev_data_))

				#		for i in range(0, 10):
				#			self.pub(_cmd, "Re-Initialize Connection")
				#			time.sleep(0.1)

				#		self.pub(_cmd, _data)
				#		logging.info("send[%s][%s] .. Done" % (_cmd, _data))
				#		return 0

				#	except Exception as e:
				#		time.sleep(1)
				#	finally:
				#		self.publock.release()
				return 0
		elif self.type_ == 1:
			self.pub_shm(_cmd, _data)

	def close(self):
		if self.type_ == 0:
			self.s.close()
			self.hb_flag = False
		elif self.type_ == 1:
			self.mmap.f.close()
			self.hb_flag = False

	@staticmethod
	def connect(self, _sock, _ip_port):
		_sock.connect(_ip_port)
		return _sock

if __name__ == "__main__":
	_h = GandanPub("127.0.0.1", 8888)
	_list = [str(a)+b+c+d+e+f+g+h+i+j for a in range(1,10)
						for b in ['e', 'f', 'g']
						for c in ['h', 'i', 'j']
						for d in ['k', 'l', 'm']
						for e in ['k', 'l', 'm']
						for f in ['k', 'l', 'm']
						for g in ['k', 'l', 'm']
						for h in ['k', 'l', 'm']
						for i in ['k', 'l', 'm']
						for j in ['k', 'l', 'm', 'a', 'b', 'c', 'd', 'e']]
	for _l in _list:
		_h.pub(sys.argv[1], _l)
		print(_l)
		import random
		import time;time.sleep(random.random()*0.1)
	_h.close()
